#include <givaro/givdegree.h>
namespace Givaro {
    const int64_t Degree::deginfty = Degree::DEGPOLYZERO;
} // Givaro
/* -*- mode: C++; tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
// vim:sts=4:sw=4:ts=4:et:sr:cino=>s,f0,{0,g0,(0,\:0,t0,+0,=s
